# Written by Bram Cohen
# see LICENSE.txt for license information

from IntraStorage import IntraStorage
from sha import sha
from threading import Event, Condition, Thread
true = 1
false = 0

def dummy_status(fractionDone = None, activity = None):
    pass

def dummy_data_flunked(size):
    pass

class StorageWrapper:
    def __init__(self, storage, request_size, hashes, 
            piece_length, finished, failed, 
            statusfunc = dummy_status, flag = Event(), check_hashes = true,
            data_flunked = dummy_data_flunked, backfunc = None):
        self.request_size = request_size
        self.hashes = hashes
        self.piece_length = piece_length
        self.finished = finished
        self.failed = failed
        self.doneflag = flag
        self.check_hashes = check_hashes
        self.data_flunked = data_flunked
        self.backfunc = backfunc
        numhashes = len(hashes)
        self.total_length = storage.get_total_length()
        self.amount_left = self.total_length
        if self.total_length <= piece_length * (numhashes - 1):
            raise ValueError, 'bad data from tracker - total too small'
        if self.total_length > piece_length * numhashes:
            raise ValueError, 'bad data from tracker - total too big'
        self.numactive = [0] * numhashes
        self.inactive_requests = [[] for i in xrange(numhashes)]
        self.total_inactive = 0
        self.have = [false] * numhashes
        self.waschecked = [check_hashes] * numhashes
        self.hasdata = [false] * numhashes
        self.accessLock = Condition()
        if len(hashes) == 0:
            finished()
            return
        self.done_checking = false

        was_allocated = storage.was_allocated()
        if was_allocated:
            self.storage = storage
            self.intrastorage = false
        else:
            self.storage = IntraStorage(storage, piece_length, flag)
            self.intrastorage = true
            if not storage.was_preexisting():
                statusfunc(activity = 'One moment please...')
                self.storage.IS_buildemptylist(statusfunc)

        if self.storage.was_preexisting():
            graphstep = int(numhashes/200)
            statusfunc(fractionDone = 0.0)
            graphnext = graphstep
            statusfunc(activity = 'checking existing data', fractionDone = 0.0)
            for i in xrange(numhashes):
                pos = i*self.piece_length
                length = min(self.piece_length, self.total_length-pos)
                if was_allocated or self.storage.isthere(pos, length):
                    self._check_single(i)
                else:
                    self._check_single(i, false)
                if flag.isSet():
                    return
                if i >= graphnext:
                    graphnext += graphstep
                    statusfunc(fractionDone = float(i+1)/numhashes)
        else:
            graphstep = int(numhashes/50)
            graphnext = graphstep
            for i in xrange(numhashes):
                self._check_single(i, false)
                if i >= graphnext:
                    graphnext += graphstep
                    statusfunc(fractionDone = float(numhashes+i+1)/(2*numhashes))
                        # start from 50%; IntraStorage.IS_buildemptylist did the first 50%

        if not was_allocated:
            self.foreground_allocate()

        statusfunc(fractionDone = 1.0)
        self.done_checking = true


    # standard routine
    def foreground_allocate(self):
        try:
            if not self.storage.IS_bump():  # run allocation cycle, if it doesn't return true...
                self.backfunc(self.foreground_allocate, 0.01) # do it again!
            else:
                self._swapstoragebackin()   # you can safely comment this out
        except IOError, e:
            self.failed('IO Error ' + str(e))

    def _swapstoragebackin(self):   # useful for diagnostics
        self.accessLock.acquire()	# dumping intrastorage isn't really necessary
        self.storage = self.storage.storage # ...but it's nice to know you can.  :-)
        self.intrastorage = false   # saves memory and processing too.
        self.accessLock.release()

    def get_amount_left(self):
        return self.amount_left

    def do_I_have_anything(self):
        return self.amount_left < self.total_length

    def _check_single(self, index, check = true):
        low = self.piece_length * index
        high = min(low + self.piece_length, self.total_length)
        length = high - low
        if check:
            self.waschecked[index] = true
            if not self.check_hashes or sha(self.storage.read(low, length)).digest() == self.hashes[index]:
                self.have[index] = true
                self.hasdata[index] = true
                self.amount_left -= length
                if self.amount_left == 0:
                    if self.intrastorage:
                        self.storage.IS_build(true)
                    self.finished()
                return
            else:
                if self.done_checking:
                    self.data_flunked(length)
        l = self.inactive_requests[index]
        x = 0
        while x + self.request_size < length:
            l.append((x, self.request_size))
            self.total_inactive += 1
            x += self.request_size
        l.append((x, length - x))
        self.total_inactive += 1

    def is_everything_pending(self):
        return self.total_inactive == 0

    def get_have_list(self):
        return self.have

    def do_I_have(self, index):
        return self.have[index]

    def do_I_have_requests(self, index):
        return self.inactive_requests[index] != []

    def new_request(self, index):
        # returns (begin, length)
        self.numactive[index] += 1
        self.total_inactive -= 1
        rs = self.inactive_requests[index]
        r = min(rs)
        rs.remove(r)
        return r

    def piece_came_in(self, index, begin, piece):
        try:
            self._piece_came_in(index, begin, piece)
        except IOError, e:
            self.failed('IO Error ' + str(e))
        except ValueError, e:
            self.failed('Error ' + str(e))

    def _piece_came_in(self, index, begin, piece):
        self.accessLock.acquire()
        self.hasdata[index] = true
        try:
            self.storage.write(index * self.piece_length + begin, piece)
            self.numactive[index] -= 1
            if (self.inactive_requests[index] == [] and 
                    self.numactive[index] == 0):
                self._check_single(index)
        finally:
            self.accessLock.release()

    def request_lost(self, index, begin, length):
        self.inactive_requests[index].append((begin, length))
        self.numactive[index] -= 1
        self.total_inactive += 1

    def get_piece(self, index, begin, length):
        try:
            return self._get_piece(index, begin, length)
        except IOError, e:
            self.failed('IO Error ' + str(e))
            return None
        except ValueError, e:
            self.failed('Error ' + str(e))
            return None

    def _get_piece(self, index, begin, length):
      self.accessLock.acquire()
      try:
        if not self.waschecked[index]:
            low = self.piece_length * index
            high = min(low + self.piece_length, self.total_length)
            if sha(self.storage.read(low, high - low)).digest() != self.hashes[index]:
                self.failed('told file complete on start-up, but piece failed hash check')
                return None
            self.waschecked[index] = true
        if not self.have[index]:
            return None
        low = self.piece_length * index + begin
        if low + length > min(self.total_length, 
                self.piece_length * (index + 1)):
            return None
        return self.storage.read(low, length)
      finally:
          self.accessLock.release()

class DummyStorage:
    def __init__(self, total, pre = false):
        self.pre = pre
        self.s = chr(0xFF) * total
        self.done = false

    def was_preexisting(self):
        return self.pre

    def get_total_length(self):
        return len(self.s)

    def read(self, begin, length):
        return self.s[begin:begin + length]

    def write(self, begin, piece):
        self.s = self.s[:begin] + piece + self.s[begin + len(piece):]

    def finished(self):
        self.done = true

def test_basic():
    ds = DummyStorage(3)
    sw = StorageWrapper(ds, 2, [sha('abc').digest()], 4, ds.finished, None)
    assert sw.get_amount_left() == 3
    assert not sw.do_I_have_anything()
    assert sw.get_have_list() == [false]
    assert sw.do_I_have_requests(0)
    x = []
    x.append(sw.new_request(0))
    assert sw.do_I_have_requests(0)
    x.append(sw.new_request(0))
    assert not sw.do_I_have_requests(0)
    x.sort()
    assert x == [(0, 2), (2, 1)]
    sw.request_lost(0, 2, 1)
    del x[-1]
    assert sw.do_I_have_requests(0)
    x.append(sw.new_request(0))
    assert x == [(0, 2), (2, 1)]
    assert not sw.do_I_have_requests(0)
    sw.piece_came_in(0, 0, 'ab')
    assert not sw.do_I_have_requests(0)
    assert sw.get_amount_left() == 3
    assert not sw.do_I_have_anything()
    assert sw.get_have_list() == [false]
    assert not ds.done
    sw.piece_came_in(0, 2, 'c')
    assert not sw.do_I_have_requests(0)
    assert sw.get_amount_left() == 0
    assert sw.do_I_have_anything()
    assert sw.get_have_list() == [true]
    assert sw.get_piece(0, 0, 3) == 'abc'
    assert sw.get_piece(0, 1, 2) == 'bc'
    assert sw.get_piece(0, 0, 2) == 'ab'
    assert sw.get_piece(0, 1, 1) == 'b'
    assert ds.done

def test_two_pieces():
    ds = DummyStorage(4)
    sw = StorageWrapper(ds, 3, [sha('abc').digest(),
        sha('d').digest()], 3, ds.finished, None)
    assert sw.get_amount_left() == 4
    assert not sw.do_I_have_anything()
    assert sw.get_have_list() == [false, false]
    assert sw.do_I_have_requests(0)
    assert sw.do_I_have_requests(1)

    assert sw.new_request(0) == (0, 3)
    assert sw.get_amount_left() == 4
    assert not sw.do_I_have_anything()
    assert sw.get_have_list() == [false, false]
    assert not sw.do_I_have_requests(0)
    assert sw.do_I_have_requests(1)

    assert sw.new_request(1) == (0, 1)
    assert sw.get_amount_left() == 4
    assert not sw.do_I_have_anything()
    assert sw.get_have_list() == [false, false]
    assert not sw.do_I_have_requests(0)
    assert not sw.do_I_have_requests(1)

    sw.piece_came_in(0, 0, 'abc')
    assert sw.get_amount_left() == 1
    assert sw.do_I_have_anything()
    assert sw.get_have_list() == [true, false]
    assert not sw.do_I_have_requests(0)
    assert not sw.do_I_have_requests(1)
    assert sw.get_piece(0, 0, 3) == 'abc'
    assert not ds.done

    sw.piece_came_in(1, 0, 'd')
    assert ds.done
    assert sw.get_amount_left() == 0
    assert sw.do_I_have_anything()
    assert sw.get_have_list() == [true, true]
    assert not sw.do_I_have_requests(0)
    assert not sw.do_I_have_requests(1)
    assert sw.get_piece(1, 0, 1) == 'd'

def test_hash_fail():
    ds = DummyStorage(4)
    sw = StorageWrapper(ds, 4, [sha('abcd').digest()], 4, ds.finished, None)
    assert sw.get_amount_left() == 4
    assert not sw.do_I_have_anything()
    assert sw.get_have_list() == [false]
    assert sw.do_I_have_requests(0)

    assert sw.new_request(0) == (0, 4)
    sw.piece_came_in(0, 0, 'abcx')
    assert sw.get_amount_left() == 4
    assert not sw.do_I_have_anything()
    assert sw.get_have_list() == [false]
    assert sw.do_I_have_requests(0)

    assert sw.new_request(0) == (0, 4)
    assert not ds.done
    sw.piece_came_in(0, 0, 'abcd')
    assert ds.done
    assert sw.get_amount_left() == 0
    assert sw.do_I_have_anything()
    assert sw.get_have_list() == [true]
    assert not sw.do_I_have_requests(0)

def test_lazy_hashing():
    ds = DummyStorage(4)
    flag = Event()
    sw = StorageWrapper(ds, 4, [sha('abcd').digest()], 4, ds.finished, lambda x, flag = flag: flag.set(), check_hashes = false)
    assert sw.get_piece(0, 0, 2) is None
    assert flag.isSet()

def test_lazy_hashing_pass():
    ds = DummyStorage(4)
    flag = Event()
    sw = StorageWrapper(ds, 4, [sha(chr(0xFF) * 4).digest()], 4, ds.finished, lambda x, flag = flag: flag.set(), check_hashes = false)
    assert sw.get_piece(0, 0, 2) is None
    assert not flag.isSet()

def test_preexisting():
    ds = DummyStorage(4, true)
    sw = StorageWrapper(ds, 2, [sha(chr(0xFF) * 2).digest(), 
        sha('ab').digest()], 2, ds.finished, None)
    assert sw.get_amount_left() == 2
    assert sw.do_I_have_anything()
    assert sw.get_have_list() == [true, false]
    assert not sw.do_I_have_requests(0)
    assert sw.do_I_have_requests(1)
    assert sw.new_request(1) == (0, 2)
    assert not ds.done
    sw.piece_came_in(1, 0, 'ab')
    assert ds.done
    assert sw.get_amount_left() == 0
    assert sw.do_I_have_anything()
    assert sw.get_have_list() == [true, true]
    assert not sw.do_I_have_requests(0)
    assert not sw.do_I_have_requests(1)

def test_total_too_short():
    ds = DummyStorage(4)
    try:
        StorageWrapper(ds, 4, [sha(chr(0xff) * 4).digest(),
            sha(chr(0xFF) * 4).digest()], 4, ds.finished, None)
        raise 'fail'
    except ValueError:
        pass

def test_total_too_big():
    ds = DummyStorage(9)
    try:
        sw = StorageWrapper(ds, 4, [sha('qqqq').digest(),
            sha(chr(0xFF) * 4).digest()], 4, ds.finished, None)
        raise 'fail'
    except ValueError:
        pass

def test_end_above_total_length():
    ds = DummyStorage(3, true)
    sw = StorageWrapper(ds, 4, [sha('qqq').digest()], 4, ds.finished, None)
    assert sw.get_piece(0, 0, 4) == None

def test_end_past_piece_end():
    ds = DummyStorage(4, true)
    sw = StorageWrapper(ds, 4, [sha(chr(0xFF) * 2).digest(), 
        sha(chr(0xFF) * 2).digest()], 2, ds.finished, None)
    assert ds.done
    assert sw.get_piece(0, 0, 3) == None
